#ifndef Controller_H
#define Controller_H

#include <iostream>
#include <sstream>
#include <list>
#include <string>

#include <boost/regex.hpp>

#include <QtGui>
//TODO Include only the necessary libraries

#include <ExtendedTask.hpp>
#include <State.hpp>
#include <Client.hpp>
#include <MsgStructure.hpp>

namespace hatptester{

class HATPController: public QObject{
	Q_OBJECT

private:
	HATPState state;
	std::map<std::string, ExtendedTask> extendedTasks;
	ExtendedTask* currentTask;
	
	QStringListModel tasksModel;
	
	HATPClient client;
	bool correctlyInitialised;
	//TODO Change localhost and port to be able to change them at the begining or at runtime
	
	int requestId;
	
	#define METHODS 0
	#define ACTIONS 1	
	//Tasktype descrobes the type of task to add to the state : either 0 for Methods or 1 for Actions
	bool initialiseTasks(bool taskType){
		std::string askTasks;
		if(taskType==METHODS){
			askTasks="{\"HATP-REQ\": { \"REQ-TYPE\": \"getTasks\"}}";
		}else{
			askTasks="{\"HATP-REQ\": { \"REQ-TYPE\": \"getActions\"}}";
		}
		std::pair<bool, std::string> tasks=client.requestValue(askTasks);
		if(tasks.first){
			boost::smatch result;
			boost::regex taskAnswerRegex;
			if(taskType==METHODS){
				taskAnswerRegex=methodAnswerRegex;
			}else{
				taskAnswerRegex=actionAnswerRegex;
			}
			if(boost::regex_match(tasks.second, result, taskAnswerRegex)){
				std::string tasks=result[1];
				boost::regex taskRegex;
				if(taskType==METHODS){
					taskRegex=methodRegex;
				}else{
					taskRegex=actionRegex;
				}
				
				boost::sregex_token_iterator taskIterator(tasks.begin(), tasks.end(), taskRegex), endTask;

				while(taskIterator != endTask){
					boost::smatch resultTask;
					const std::string taskToProcess=*taskIterator;
					
					if(boost::regex_match(taskToProcess, resultTask, taskRegex)){
						std::string parameters=resultTask[2];
						boost::sregex_token_iterator parameterIterator(parameters.begin(), parameters.end(), wordRegex), endParameter;
						Task::Parameters temp;
						unsigned long int i=0;
						while(parameterIterator != endParameter){
							std::stringstream tempName; //TODO replace by actual name from domain
							tempName<<*parameterIterator<<i;
							temp.push_back(Task::Parameter(*parameterIterator, tempName.str()));
							++parameterIterator;
							++i;
						}
		
						if(taskType==METHODS){
							state.addMethod(Method(resultTask[1], temp));
						}else{
							state.addAction(Action(resultTask[1], temp));
						}
					}
					++taskIterator;
				}
				
				return true;
			}
		}else{
			client.stopThread();
		}
		
		return false;
	}
	
	bool initialiseEntities(){
		bool success=false;
		std::string askEntities="{\"HATP-REQ\": { \"REQ-TYPE\": \"getEntities\"}}";
		std::pair<bool, std::string> entities=client.requestValue(askEntities);
		if(entities.first){
			boost::smatch result;
			if(boost::regex_match(entities.second, result, entityAnswerRegex)){
				success=true;
				std::string entities=result[1];
								
				boost::sregex_token_iterator entitiesIterator(entities.begin(), entities.end(), entityRegex), endEntities;

				while(entitiesIterator != endEntities){
					boost::smatch resultEntity;
					const std::string entityToProcess=*entitiesIterator;
					if(boost::regex_match(entityToProcess, resultEntity, entityRegex)){
						if(resultEntity.size()!=3){
							success=false;
						}
						state.addEntity(Entity(resultEntity[2], resultEntity[1]));
						success=success and true;
					}
					++entitiesIterator;
				}
			}
		}else{
			client.stopThread();
		}
		
		return success;
	}
	
	HATPController():client("HATPControl", "localhost", 5500), correctlyInitialised(false){
		if(client.isConnected()){
			bool success;
			//Initialise state using requests to HATP
			success=initialiseEntities();

			if(success){
				success=initialiseTasks(METHODS); //Get Methods
				
				if(success){				
					success=initialiseTasks(ACTIONS); //Get Actions
					
					if(success){		
						//Initialise the different models for Qt GUI
						initialiseTaskModel();
					
						currentTask=NULL;
						requestId=0;
						correctlyInitialised=true;
					}
				}
			}
		}
	}
	HATPController(HATPController const&);
	HATPController& operator=(HATPController const&);
	
	void initialiseTaskModel(){
		QStringList tasks;
		//TODO Remove those marking and put unselectable labels instead (with different backgroung color)
		tasks << " -- Methods : ";
		
		for(std::map<std::string, Method>::const_iterator it=state.getMethods().begin(); it!=state.getMethods().end(); ++it){
			tasks << it->first.c_str();
			extendedTasks.insert(std::make_pair(it->first, ExtendedTask(state, it->first)));
		}
		tasks << "";

		tasks << " -- Actions : ";
		for(std::map<std::string, Action>::const_iterator it=state.getActions().begin(); it!=state.getActions().end(); ++it){
			tasks << it->first.c_str();
			extendedTasks.insert(std::make_pair(it->first, ExtendedTask(state, it->first)));
		}

		tasksModel.setStringList(tasks);
	}
	
	void exit(){
		client.stopThread();
		emit sendStop();
	}
	
public:	
	~HATPController(){
		client.stopThread();
	}
	
	bool isReady(){
		return(client.isConnected()&&correctlyInitialised);
	}
	
	const std::map<std::string, Entity> getEntities() const{
		return(state.getEntities());
	}
	
	void sendRequest(){
		std::stringstream request;
		
		//Header
		request<<"{\"HATP-REQ\": { \"REQ-TYPE\": \"plan\", \"REQ-ID\": "<<requestId<<", ";
		++requestId;
		
		//Task
		request<<"\"TASK-NAME\": \""<<currentTask->getName()<<"\", ";
		
		//Parameters
		request<<"\"PARAMETERS\": [";
		std::pair<bool, const Task*> result;
		result=state.getTask(currentTask->getName());
		if(!result.first){
			return;
		}
		const Task::Parameters& parameters=result.second->getParameters();
		size_t index=0;
		for(Task::Parameters::const_iterator it=parameters.begin(); it!=parameters.end(); ++it){
			++index;
			request<<"\""<<it->value<<"\"";
			if(index!=parameters.size()){
				request<<", ";
			}			
		}
		
		//Footer
		request<<"]}}";
		
		//Send the request
		if(!client.sendMessage(request.str())){
			client.stopThread();
		}
	}
	
	QStringListModel& getTasks(){
		return tasksModel;
	}
	
	QStandardItemModel* getTaskParameter(const std::string& task){
		std::map<std::string, ExtendedTask>::iterator result;
		result=extendedTasks.find(task);
		if(result!=extendedTasks.end()){
			currentTask=&result->second;
			return result->second.getParametersModel();
		}
		return NULL;
	}
	
	bool isReadyTask(const std::string& task){
		std::pair<bool, const Task*> result;
		result=state.getTask(task);
		if(result.first){
			return result.second->isReady();
		}
		return false;
	}
	
	bool isReadyCurrentTask(){
		std::pair<bool, const Task*> result;
		result=state.getTask(currentTask->getName());
		if(result.first){
			return result.second->isReady();
		}
		return false;
	}
	
	const std::string getCurrentTaskName() const{
		return currentTask->getName();
	}
	
	std::set<std::string> getMissingParameters(const std::string& task){
		std::set<std::string> missingParameters;
	
		std::pair<bool, const Task*> result;
		result=state.getTask(task);
		if(result.first && !result.second->isReady()){
			const Task::Parameters& parameters=result.second->getParameters();
			for(Task::Parameters::const_iterator it=parameters.begin(); it!=parameters.end(); ++it){
				if(it->value==""){
					missingParameters.insert(it->name+" ["+it->type+"]");
				}
			}
		}
		return missingParameters;
	}
	
	const std::string getLongestTaskName() const{
		std::string longest="";
		for(HATPState::MethodsMap::const_iterator it=state.getMethods().begin(); it!=state.getMethods().end(); ++it){
			if(it->first.length()>longest.length()){
				longest=it->first;
			}
		}
		for(HATPState::ActionsMap::const_iterator it=state.getActions().begin(); it!=state.getActions().end(); ++it){
			if(it->first.length()>longest.length()){
				longest=it->first;
			}
		}
		return longest;
	}

	static HATPController& getInstance(){
		static HATPController instance;
		return instance;
	}
	
signals:
	void sendStop();
};

}; //namespace

#endif //Controller_H;
